/*
	Simple udp server
*/
#include<stdio.h>	//printf
#include<string.h> //memset
#include <unistd.h> //sleep();
#include<stdlib.h> //exit(0);
#include<arpa/inet.h>
#include<sys/socket.h>
#include <time.h>
#include <pthread.h>
#include <sys/prctl.h>

#define SECOND  1000000000uL
#define MSEC 1000000uL

#define PORT 8090	//The port on which to send data
#define PORT_HANDSHAKE  8070

#define OBJ1_SIZE        (80*1024)
#define OBJ2_SIZE        (10*1024)
#define BUFFER_SIZE     (100*1024)

#define PAYLOAD_SIZE    1440
#define HEADER_SIZE       18
#define PACKET_SIZE (PAYLOAD_SIZE + HEADER_SIZE)
#define MAX_PACKETS_PER_CHUNK   10
#define MAX_CHUNK_SIZE  (PAYLOAD_SIZE*MAX_PACKETS_PER_CHUNK)

#define NUMBER_OF_SAMPLES   10
#define NUMBER_OF_SAMPLES2  20
#define MAX_NUMBER_OF_SAMPLES   100000

/* Dummy data array containing payload and MW Header */
static uint8_t _dummy_data[BUFFER_SIZE];

/* Target address struct */
static struct sockaddr_in _si_other;
/* Socket */
static int _socket;

typedef struct {
    uint16_t seqNr;
    uint16_t fragNr;
    struct timespec timestamp;
    struct timespec chunk_timestamp;
    int lastFrag;
} t_dummy_header;

static long _latencies[MAX_NUMBER_OF_SAMPLES][100];

static int _isMaster = 0;

static int _socket_prio = 0;


void die(char *s){
	perror(s);
	exit(1);
}

void send_fragment(int seqNr, int fragNr, int fragSize, struct timespec *startTime){
    t_dummy_header *header = (t_dummy_header*) (_dummy_data + fragNr*PACKET_SIZE);
    header->seqNr = seqNr;
    header->fragNr = fragNr;
    //header->lastFrag = lastFrag;
    memcpy(&header->chunk_timestamp, startTime, sizeof(struct timespec));
    clock_gettime ( CLOCK_REALTIME , &header->timestamp);
      
    if (sendto(_socket, (void*)header, fragSize , 0 , (struct sockaddr *) &_si_other, sizeof(_si_other))==-1)
        die("sendto()");
}

void send_chunk(int seqNr, int chunkNr, struct timespec *startTime, int obj_size){
    int bytesLeft = obj_size - (MAX_CHUNK_SIZE * chunkNr);
    int i = 0;
   
    if(bytesLeft >= MAX_CHUNK_SIZE){
        /* Send a whole chunk */
        for(i = 0; i < MAX_PACKETS_PER_CHUNK; i++){
            send_fragment(seqNr, (chunkNr*MAX_PACKETS_PER_CHUNK)+i, PACKET_SIZE, startTime);
        }
    } else {
        /* This chunk is the last one and smaller */
        int fullSizePacketsInChunk = bytesLeft / PAYLOAD_SIZE;
        for(i = 0; i < fullSizePacketsInChunk; i++){
            send_fragment(seqNr, (chunkNr*MAX_PACKETS_PER_CHUNK)+i, PACKET_SIZE, startTime);
            bytesLeft -= PAYLOAD_SIZE;
        }
        if(bytesLeft > 0){
            send_fragment(seqNr, (chunkNr*MAX_PACKETS_PER_CHUNK)+i, bytesLeft, startTime);
        }
    }
}

static void _t_add(struct timespec *t, long nsec){
    t->tv_nsec += nsec;
    if(t->tv_nsec >= SECOND){
        t->tv_nsec -= SECOND;
        t->tv_sec++;
    }
}

static long _t_diff(struct timespec *t1, struct timespec *t2){
    struct timespec diff;
    if (t2->tv_nsec-t1->tv_nsec < 0) {
        diff.tv_sec  = t2->tv_sec - t1->tv_sec - 1;
        diff.tv_nsec = t2->tv_nsec - t1->tv_nsec + 1000000000;
    } else {
        diff.tv_sec  = t2->tv_sec - t1->tv_sec;
        diff.tv_nsec = t2->tv_nsec - t1->tv_nsec;
    }
    return (diff.tv_sec * 1000000000.0 + diff.tv_nsec);
}

void _print_latencies(int isMaster, int number_of_hyperperiods){
    FILE *fptr;
    
    if(isMaster)
        fptr = fopen("latencies.csv", "w");
    else
        fptr = fopen("latencies2.csv", "w");
    
    if(fptr == NULL)
        die("fopen");
    
    int nr_of_samples = isMaster ? number_of_hyperperiods*2 : number_of_hyperperiods;
    
    for(int i = 0; i < nr_of_samples; i++){
        for(int j = 0; j < 100; j++){
            //if(_latencies[i][j] != 0){
                fprintf(fptr, "%d,", _latencies[i][j]);
            //}
        }
        fprintf(fptr,"\n");
    }
    fclose(fptr);
}

void *rx_thread_function(void *p_arg){
    int rx_socket, recv_len;
    struct sockaddr_in si_me, si_other;
    int slen = sizeof(si_other);
    char buf[PACKET_SIZE];
    struct timespec t_now;
    int *isMaster = (int*)p_arg;
    
    printf("starting rx thread as %d\n", *isMaster);
    
    if((rx_socket=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
		die("rxsocket");
	
	if(setsockopt(rx_socket, SOL_SOCKET, SO_PRIORITY, &_socket_prio, sizeof(int)) == -1)
		die("setsockopt");
    
    // zero out the structure
	memset((char *) &si_me, 0, sizeof(si_me));
	
	si_me.sin_family = AF_INET;
    si_me.sin_port = (*isMaster)==1 ? htons(PORT+1) : htons(PORT);
	si_me.sin_addr.s_addr = htonl(INADDR_ANY);
	
	//bind socket to port
	if(bind(rx_socket, (struct sockaddr*)&si_me, sizeof(si_me)) == -1)
		die("bind");
    
    while(1){
        if ((recv_len = recvfrom(rx_socket, buf, PACKET_SIZE, 0, (struct sockaddr *) &si_other, &slen)) == -1)
			die("recvfrom()");
        clock_gettime(CLOCK_REALTIME, &t_now);
        
        t_dummy_header *header = (t_dummy_header*)buf;
        _latencies[header->seqNr][header->fragNr] = _t_diff(&header->chunk_timestamp, &t_now);
    }
       
    return NULL;
}

void _do_handshake(int isMaster, struct timespec *t_now){
    int s, recv_len;
    struct sockaddr_in si_me, si_other, si_tmp;
    int slen = sizeof(si_other);
    char buf[PACKET_SIZE];
    
    /* set timer slack to 1 ns */
    prctl( PR_SET_TIMERSLACK , 1);
    
    if((s=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
		die("rxsocket");
		
	if(setsockopt(s, SOL_SOCKET, SO_PRIORITY, &_socket_prio, sizeof(int)) == -1)
		die("setsockopt");
    
    // zero out the structure
	memset((char *) &si_me, 0, sizeof(si_me));
	
	si_me.sin_family = AF_INET;
    si_me.sin_port = isMaster==1 ? htons(PORT_HANDSHAKE+1) : htons(PORT_HANDSHAKE);
	si_me.sin_addr.s_addr = htonl(INADDR_ANY);
    
    memcpy(&si_other, &_si_other, sizeof(struct sockaddr_in));
    si_other.sin_port = isMaster==1 ? htons(PORT_HANDSHAKE) : htons(PORT_HANDSHAKE+1);
	
	//bind socket to port
	if(bind(s, (struct sockaddr*)&si_me, sizeof(si_me)) == -1)
		die("bind");
    
    if(isMaster == 1){
        /* We are master */
        /* Send sync */
        buf[0] = 0xAB;
        if (sendto(s, (void*)buf, 1 , 0 , (struct sockaddr *) &si_other, sizeof(si_other))==-1)
            die("sendto()");
        /* Wait for reply */
        if ((recv_len = recvfrom(s, buf, 1, 0, (struct sockaddr *) &si_tmp, &slen)) == -1)
			die("recvfrom()");
        clock_gettime(CLOCK_REALTIME , t_now);
        t_now->tv_nsec=0;
        t_now->tv_sec = t_now->tv_sec+5;
        /* send start time to slave */
        memcpy(buf,t_now,sizeof(struct timespec));
        if (sendto(s, (void*)buf, sizeof(struct timespec) , 0 , (struct sockaddr *) &si_other, sizeof(si_other))==-1)
            die("sendto()");
    } else {
        /* We are slave */
        /* Wait for sync */
        if ((recv_len = recvfrom(s, buf, 1, 0, (struct sockaddr *) &si_tmp, &slen)) == -1)
			die("recvfrom()");
        /* Reply */
        buf[0]++;
        if (sendto(s, (void*)buf, 1 , 0 , (struct sockaddr *) &si_other, sizeof(si_other))==-1)
            die("sendto()");
        /* Wait for start time */
        if ((recv_len = recvfrom(s, buf, sizeof(struct timespec), 0, (struct sockaddr *) &si_tmp, &slen)) == -1)
			die("recvfrom()");
        memcpy(t_now, buf, sizeof(struct timespec));
    }
    
    printf("Will start at %d s; %d ns\n",t_now->tv_sec, t_now->tv_nsec);
    /* wait until start time */
    clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, t_now, NULL);
    
}

int main(int argc, char *argv[]){	
    struct timespec t_now, t_add;
    int err = 0;
    t_add.tv_sec = 0;
    int number_of_hyperperiods = 0;
    
    
    if(argc < 4 || argc > 5)
        die("Invalid arguments");
        
    if(argc == 5)
		_socket_prio = atoi(argv[4]);
        
    _isMaster = atoi(argv[2]);
    number_of_hyperperiods = atoi(argv[3]);
    
    pthread_t rx_thread;
    
    pthread_create(&rx_thread, NULL, rx_thread_function, (void*)&_isMaster);
    
    /* Initialize dummy data and calculate number of fragments to transfer */
    memset(_dummy_data, 0x0, sizeof(_dummy_data));
    
    if((_socket=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
		die("socket");
		
	if(setsockopt(_socket, SOL_SOCKET, SO_PRIORITY, &_socket_prio, sizeof(int)) == -1)
		die("setsockopt");

	memset((char *) &_si_other, 0, sizeof(_si_other));
	_si_other.sin_family = AF_INET;
    _si_other.sin_port = (_isMaster)==1 ? htons(PORT) : htons(PORT+1);
	if (inet_aton(argv[1] , &_si_other.sin_addr) == 0) 
        die("inet_aton() failed\n");
    
    _do_handshake(_isMaster, &t_now);

    if(_isMaster){
        /* Send 80kB sample each 10ms */
        for(int seqNr = 0; seqNr < number_of_hyperperiods+1; seqNr++){
            for(int chunkNr = 0; chunkNr < 6; chunkNr++){
                send_chunk(seqNr, chunkNr, &t_now, OBJ1_SIZE);
                _t_add(&t_now, MSEC);
                clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &t_now, NULL);
            }
            _t_add(&t_now, 4*MSEC);
            clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &t_now, NULL);
        }
    } else {
        /* Send 10kB sample each 5ms */
        for(int seqNr = 0; seqNr < (number_of_hyperperiods*2)+1; seqNr++){
            send_chunk(seqNr, 0, &t_now, OBJ2_SIZE);
            _t_add(&t_now, 5*MSEC);
            clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &t_now, NULL);
        }
    }

    _print_latencies(_isMaster, number_of_hyperperiods);

	return 0;
}
